##     Copyright (C) 2003 Christian Kristukat (ckkart@hoc.net)     

##     This program is free software; you can redistribute it and/or modify
##     it under the terms of the GNU General Public License as published by
##     the Free Software Foundation; either version 2 of the License, or
##     (at your option) any later version.

##     This program is distributed in the hope that it will be useful,
##     but WITHOUT ANY WARRANTY; without even the implied warranty of
##     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##     GNU General Public License for more details.

##     You should have received a copy of the GNU General Public License
##     along with this program; if not, write to the Free Software
##     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import numpy as np
    
from scipy import special
from scipy.integrate import trapz

from peak_o_mat import peaksupport as ps

def lbstep(x,amp,pos,sigma,sigmap=100.0,rng=5,pts=1000):
    """\
lorentian broadened step function according to
http://www.pfk.ff.vu.lt/Mokslas/publ_pdf/2007_Quantum%20Well%20disorder%20effects.pdf

caution: this function implies numerical convolution which is
computationally intensive.  For accurate results you should run
convergence tests taking into account the 'rng' and 'pts' parameters
in the function definition in peaks.py.

    """ 
    if len(x) == 1:
        dx = x[0]*0.1
    else:
        dx = x[-1]-x[0]
    up = x[-1]+rng*dx
    lo = x[0]-rng*dx
    f = amp/np.pi
    xp = np.linspace(lo,up,pts)[:,np.newaxis]
    s2 = np.power(sigma,2)
    i = trapz(sigma/(np.power(x-xp,2)+s2)/(1+np.exp(-sigmap*(xp-pos))), xp, axis=0)
    return f*i

def gbstep(x,amp,pos,sigma,sigmap=100,rng=5,pts=1000):
    """\
gaussian broadened step function according to
http://www.pfk.ff.vu.lt/Mokslas/publ_pdf/2007_Quantum%20Well%20disorder%20effects.pdf

caution: this function implies numerical convolution which is
computationally intensive.  For accurate results you should run
convergence tests taking into account the 'rng' and 'pts' parameters
in the function definition in peaks.py.

    """
    f = amp/np.sqrt(2*np.pi)/sigma
    s2 = 2*np.power(sigma,2)
    if len(x) == 1:
        dx = x[0]*0.1
    else:
        dx = x[-1]-x[0]
    up = x[-1]+rng*dx
    lo = x[0]-rng*dx
    xp = np.linspace(lo,up,pts)[:,np.newaxis]
    i = trapz(np.exp(-np.power(x-xp,2)/s2)/(1+np.exp(-sigmap*(xp-pos))), xp, axis=0)
    return f*i

def voigt(x,amp,pos,fwhm,shape):
    """\
voigt profile

V(x,sig,gam) = Re(w(z))/(sig*sqrt(2*pi))
z = (x+i*gam)/(sig*sqrt(2))
    """
    tmp = 1/special.wofz(np.zeros((len(x))) +1j*np.sqrt(np.log(2.0))*shape).real
    return tmp*amp*special.wofz(2*np.sqrt(np.log(2.0))*(x-pos)/fwhm+1j*np.sqrt(np.log(2.0))*shape).real

def sdec(x, amp, tau, beta):
    """\
stretched exponential decay
    """
    y = amp*np.exp(-np.power(x/tau,beta))*(x>=0)
    tmp = np.zeros(x.shape,dtype=float)
    mask = np.compress(np.isfinite(y),np.arange(len(y),dtype=int))
    np.put(tmp, mask, np.compress(np.isfinite(y),y))
    return tmp

ps.add('CB',
    func='const+0*x',
    info='constant background', 
    picker=ps.CBPicker, 
    ptype='BACKGROUND')

ps.add('LB',
    func='const+lin*x',
    info='linear background', 
    picker=ps.LBPicker, 
    ptype='BACKGROUND')

ps.add('QB',
    func='a*x**2+b*x+c',
    info='quadratic background', 
    picker=ps.QBPicker, 
    ptype='BACKGROUND')

ps.add('GA',
    func='amp*np.exp(-(np.power(x-pos,2)/(fwhm*fwhm/4.0/np.log(2.0))))',
    info='symmetric gaussian', 
    picker=ps.LOPicker,
    ptype='PEAK')

ps.add('LO',
    func='amp/(1+np.power((x-pos)/(fwhm/2.0),2))',
    info='symmetric lorentzian', 
    picker=ps.LOPicker, 
    ptype='PEAK')

ps.add('FAN',
    func='amp*np.power((1+(x-pos)/(fwhm/2)*shape),2)/(1+np.power((x-pos)/(fwhm/2),2))',
    info='fano lineshape', 
    picker=ps.FANPicker, 
    ptype='PEAK')

ps.add('VO',
    func='voigt(x,amp,pos,fwhm,shape)',
    info=voigt.__doc__, 
    picker=ps.FANPicker, 
    ptype='PEAK')

ps.add('PSV',
    func='amp*(shape*np.exp(-2/np.pi*np.power((x-pos)/(fwhm/2),2))+(1-shape)/(1+np.power((x-pos)/(fwhm/2),2)))',
    info='pseudo voigt profile', 
    picker=ps.FANPicker, 
    ptype='PEAK')

ps.add('DEC',
    func='amp*np.exp(-x/tau)',
    info='exponential decay',
    ptype='EXP',
    picker=ps.EXPPicker)

ps.add('RISE',
    func='1.0-np.exp(-x/tau)',
    info='exponential rise', 
    ptype='EXP')

ps.add('SDEC',
    func='sdec(x,amp,tau,beta)',
    info='stretched exponential decay', 
    ptype='EXP',
    picker=ps.SEXPPicker)

ps.add('LBSTP',
    info=lbstep.__doc__,
    func='lbstep(x, amp, pos, sigma)',
    picker=ps.STEPPicker,
    ptype='STEP')

ps.add('GBSTP',
    info=gbstep.__doc__,
    func='gbstep(x, amp, pos, sigma)',
    picker=ps.STEPPicker,
    ptype='STEP')

ps.add('FERMI', 
    info='fermi function with abscissa in eV',
    func='amp/(np.exp(-(x-energy)/(c_kb/c_e*T))+1)',
    picker=ps.FERMIPicker,
    ptype='STEP')

